/*
 * ManagerImpl.hh
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file ManagerImpl.hh
 *
 * Header file for the default implementation of the Broker::Manager interface.
 */

#ifndef _manager_impl_hh
#define _manager_impl_hh

#include "BrokerS.h"

#include <listNode.h>

/**
 * A default implementation of the Broker::Manager interface.
 */
class ManagerImpl : public virtual POA_Broker::Manager
{

public:

    /**
     * Construct a ManagerImpl.
     *
     * @param name The name of this manager.
     */
    ManagerImpl(const char *name);

    /**
     * Destruct a ManagerImpl.
     */
    virtual ~ManagerImpl();

    /** @copydoc Broker::Manager::CurrentPolicy */
    virtual Broker::Policy_ptr CurrentPolicy(void)
	throw (CORBA::SystemException);
    
    /** @copydoc Broker::Manager::CurrentPolicy */
    virtual void CurrentPolicy(Broker::Policy_ptr policy)
	throw (CORBA::SystemException);

    /** @copydoc Broker::Manager::AddTask */
    virtual void AddTask(Broker::Task_ptr task,
			 const Broker::ScheduleParameters &sp)
	throw (CORBA::SystemException,
	       Broker::DuplicateScheduleParameter,
	       Broker::InvalidScheduleParameter,
	       Broker::MissingScheduleParameter);

    /** @copydoc Broker::Manager::RemoveTask */
    virtual void RemoveTask(Broker::Task_ptr task)
	throw (CORBA::SystemException);

    /** @copydoc Broker::Manager::GetTaskList */
    virtual Broker::TaskList *GetTaskList(void)
	throw (CORBA::SystemException);

    /** @copydoc Broker::Manager::ChangeTaskCPU */
    virtual void ChangeTaskCPU(Broker::RealTimeTask_ptr task,
			       CORBA::ULong ct,
			       CORBA::ULong status,
			       CORBA::ULong advice)
	throw (CORBA::SystemException, Broker::InvalidState);

    /**
     * Search of a task name in a Broker::TaskList.  Or, if the name is an IOR
     * return a reference to that object.
     *
     * @param orb A pointer to an active CORBA::ORB object.
     * @param tl The task list to search.
     * @param name The task name or task IOR.
     * @return An object reference that matches the given name or IOR.
     * @exception CORBA::BAD_PARAM if the task could not be found.
     */
    static Broker::Task_ptr ResolveTask(CORBA::ORB_ptr orb,
					Broker::TaskList &tl,
					const char *name)
	throw (CORBA::SystemException);
    
private:

    /**
     * Repair the task list by removing any dead/unreachable objects.
     */
    void RepairTaskList(void);

    /**
     * Get the index of the given task in the mi_Tasks sequence.
     *
     * @param task The task object to look up.
     * @return The index of the given task in the mi_Tasks sequence or -1 if
     *         the task is not in the sequence.
     */
    int TaskIndex(Broker::Task_ptr task);
    
    /**
     * Remove a task object at a given index.
     *
     * @param index The index in mi_Tasks to remove.
     */
    void RemoveTaskAt(unsigned int index);

    /**
     * The active policy object or nil.
     */
    Broker::Policy_var mi_CurrentPolicy;

    /**
     * The list of tasks currently being managed.  This list will be passed
     * to policies when they are Broker::Policy::Activate()'d.
     */
    Broker::TaskList mi_Tasks;
    
};

#endif
