/*
 * NamingHelper_T.h
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
/*  -*- C++ -*- */

//==========================================================================
/**
 * @file     NamingHelper_T.h
 *
 * NamingHelper_T.h,v 1.4 2003/11/20 17:13:44 stack Exp
 *
 * @author  Craig Rodrigues (crodrigu@bbn.com)
 */
//==========================================================================
                             

#ifndef NAMING_HELPER_T_H
#define NAMING_HELPER_T_H

#include "CosNamingC.h"

/**
  * @class NamingHelper
  *
  * @brief  Utility functions for using the CORBA Naming Service.
  *
  * This class uses template functions outlined in Chapter 18
  * of "Advanced CORBA Programming with C++" by Michi Henning and
  * Steve Vinoski.
  */
template<class T>
class NamingHelper
{
public:
  /// template function for resolve_initial_references()
  static
  typename T::_ptr_type
  resolve_init(CORBA::ORB_ptr orb, const char *id,
               size_t tries = 1,
               ACE_Time_Value timeout = ACE_Time_Value::zero)
    throw ( CORBA::SystemException,
	    CORBA::ORB::InvalidName
            );

  /// template function for resolving a name in the CORBA Naming Service
  static
  typename T::_ptr_type
  resolve_name( CosNaming::NamingContext_ptr nc,
                const CosNaming::Name& name,
                size_t tries = 1,
                ACE_Time_Value timeout = ACE_Time_Value::zero)
    throw( CORBA::SystemException,
           CosNaming::NamingContext::NotFound,
           CosNaming::NamingContext::CannotProceed,
           CosNaming::NamingContext::InvalidName); 

  /// template function to bind a name to the CORBA Naming Service
  /** template function for binding to the CORBA Naming Service,
   * with tries and a timeout value
   */
  static
  void bind( CosNaming::NamingContext_ptr nc,
             const CosNaming::Name& n, CORBA::Object_ptr obj,
             size_t tries = 1, 
             ACE_Time_Value timeout = ACE_Time_Value::zero)
    throw ( CORBA::SystemException,
            CosNaming::NamingContext::NotFound,
            CosNaming::NamingContext::CannotProceed,
            CosNaming::NamingContext::InvalidName,
            CosNaming::NamingContext::AlreadyBound );

  /// template function to rebind a name to the CORBA Naming Service
  /** template function for rebinding to the CORBA Naming Service,
   * with tries and a timeout value
   */
  static
  void rebind ( CosNaming::NamingContext_ptr nc,
                const CosNaming::Name & n, CORBA::Object_ptr obj,
                size_t tries = 1, 
                ACE_Time_Value timeout = ACE_Time_Value::zero)
    throw (
           CORBA::SystemException,
           CosNaming::NamingContext::NotFound,
           CosNaming::NamingContext::CannotProceed,
           CosNaming::NamingContext::InvalidName
           );                                     


};


template<class T>
typename T::_ptr_type
NamingHelper<T>::resolve_init(CORBA::ORB_ptr orb, const char *id,
                              size_t tries,
                              ACE_Time_Value timeout)
  throw ( CORBA::SystemException,
          CORBA::ORB::InvalidName
          )
{
  size_t i;
  CORBA::Object_var obj;
  for(i=0; i < tries; i++ ){
    try{
      obj = orb->resolve_initial_references( id );
      break;  // we resolved a name without throwing an exception
      // so break out of this for loop
    }
    catch( ... ){
      if( i < tries ){
        ACE_DEBUG((LM_DEBUG, "Retrying resolve() of %s ...%d\n", id, i));

        // ACE_OS::sleep() does a select() with a timeout, so is
        // threadsafe
        ACE_OS::sleep(timeout);
      }else{
        // We have exceeded our tries, rethrowing exception
        ACE_DEBUG((LM_DEBUG, "Rethrowing exception\n"));
        throw;
      }
    }
  } 
  //assert( !CORBA::is_nil(obj.in() ));

  typename T::_var_type ref;
  try{
    ref = T::_narrow(obj.in() );
  }catch( const CORBA::SystemException& e){
    throw;
  }
	      
  return ref._retn(); 
}


template< class T>
typename T::_ptr_type
NamingHelper<T>::resolve_name( CosNaming::NamingContext_ptr nc,
                               const CosNaming::Name& name, 
                               size_t tries,
                               ACE_Time_Value timeout)
  throw( CORBA::SystemException,
         CosNaming::NamingContext::NotFound,
         CosNaming::NamingContext::CannotProceed,
         CosNaming::NamingContext::InvalidName 
         )
{
  CORBA::Object_var obj;
  size_t i;

  for(i=0; i < tries; i++ ){
    try{
      obj = nc->resolve( name );
      break;  // we resolved a name without throwing an exception
      // so break out of this for loop
    }
    catch( ... ){
      if( i < tries ){
        ACE_DEBUG((LM_DEBUG, "Retrying resolve() of %s ...%d\n", name[0].id.in(), i));

        // ACE_OS::sleep() does a select() with a timeout, so is
        // threadsafe
        ACE_OS::sleep(timeout);
      }else{
        // We have exceeded our tries, rethrowing exception
        ACE_DEBUG((LM_DEBUG, "Rethrowing exception\n"));
        throw;
      }
    }
  } 
  if( CORBA::is_nil( obj.in() )){
    // do something

  }

  typename T::_var_type ref;
  try{
    ref = T::_narrow(obj.in() );
  }catch(const CORBA::SystemException& e){
    e._tao_print_exception("");
    throw;
  }	     

  if( CORBA::is_nil(ref.in() )){
    // Do something
         
  }

  return ref._retn();
}

template< class T>
void
NamingHelper<T>::bind( CosNaming::NamingContext_ptr nc,
                       const CosNaming::Name& n, CORBA::Object_ptr obj,
                       size_t tries, ACE_Time_Value timeout)
  throw ( CORBA::SystemException,
          CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          CosNaming::NamingContext::AlreadyBound )
{
  
  size_t i;

  ACE_DEBUG((LM_DEBUG, "Calling bind() with %d tries\n", tries));
  for(i = 0; i < tries; i++)
    {
      try
        {
          nc->bind(n, obj);  // bind the name

          break;     // an exception was not thrown,
          // success, get out of this loop
	}catch(const CORBA::SystemException& e){
	  ACE_DEBUG((LM_DEBUG, "bind() threw exception\n"));
	  e._tao_print_exception("");
          // binding threw an exception, try to recover
          if(i < tries)
            {
              // sleep, and try again in the next iteration
              // of this loop.  ACE_OS::sleep() does a select()
              // on a timeout, therefore is threadsafe
              ACE_OS::sleep(timeout);
              ACE_DEBUG((LM_DEBUG, "Retrying bind() of %s...%d\n",
			 n[0].id.in(), i));
            }    
          else{
            // we have exceed our tries,  rethrowing
            ACE_DEBUG((LM_DEBUG, "Rethrowing bind() exception\n"));
            throw;
          } 
        } catch (const CORBA::Exception& e) {
	  ACE_DEBUG((LM_DEBUG, "bind() threw non-system exception\n"));
	  e._tao_print_exception("");
	  throw;
	}
    }
}

template< class T>
void
NamingHelper<T>::rebind( CosNaming::NamingContext_ptr nc,
                         const CosNaming::Name& n, CORBA::Object_ptr obj,
                         size_t tries, ACE_Time_Value timeout)
  throw ( CORBA::SystemException,
          CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName)
{
  
  size_t i;

  for(i = 0; i < tries; i++)
    {
      try
        {
          nc->rebind(n, obj);  // rebind the name

          break;     // an exception was not thrown,
          // success, get out of this loop
        }catch(...){
          // rebinding threw an exception, try to recover
          if(i < tries)
            {
              // sleep, and try again in the next iteration
              // of this loop.  ACE_OS::sleep() does a select()
              // on a timeout, therefore is threadsafe
              ACE_OS::sleep(timeout);
              ACE_DEBUG((LM_DEBUG, "Retrying...%d\n", i));
            }    
          else{
            // we have exceed our tries,  rethrowing
            ACE_DEBUG((LM_DEBUG, "Rethrowing rebind() exception\n"));
            throw;
          } 
        }
    }
}

#endif   /* NAMING_HELPER_T_H */

